/*****************************************************************************
* Chuong trinh C5-2: Chuong trinh ken chon pho ma theo giai thuat Josephus   *
*                    Chuong trinh cai dat danh sach lien ket vong dung bien  *
*                    dong, plist la con tro chi nut cuoi danh sach lien ket  *
*                    vong                                                    *
* Nguoi viet chuong trinh: Nguyen Hong Chuong                                *
* Ngay cap nhat moi nhat: 15/3/1999                                          *
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>

#define TRUE 1
#define FALSE 0

// Khai bao mau tin mo ta hoang tu
typedef struct hoangtu
{
   char ten[20];
};

// Khai bao cau truc cua mot nut tren dslk vong
struct node
{
   hoangtu info;
   struct node *next;
};

// Khai bao kieu con tro chi nut
typedef struct node *NODEPTR;

// Tac vu getnode: cap phat mot nut cho danh sach lien ket vong
NODEPTR getnode(void)
{
   NODEPTR p;
   p = (NODEPTR)malloc(sizeof(struct node));
   return(p);
}

// Tac vu freenode: huy nut da cap phat
void freenode(NODEPTR p)
{
   free(p);
}

// Tac vu initialize: khoi dong danh sach lien ket vong
void initialize(NODEPTR *plist)
{
   *plist = NULL;
}

// Tac vu empty: kiem tra dslk vong co bi rong khong
int empty(NODEPTR *plist)
{
   return((*plist == NULL) ? TRUE : FALSE);
}

// Tac vu listsize: xac dinh so nut hien co trong danh sach lien ket vong
int listsize(NODEPTR *plist)
{
   NODEPTR p;
   int i;
   if(empty(plist))
      return(0);
   p = (*plist)->next;
   i = 1;
   while(p != *plist)
   {
      i++;
      p = p->next;
   }
   return(i);
}

/* Ham nodepointer: tra ve con tro cua nut thu i (i=0,1,2,...) trong
   danh sach lien ket vong */
NODEPTR nodepointer(NODEPTR *plist, int i)
{
   NODEPTR p;
   int vitri;
   if(i < 0 || i >= listsize(plist))
      return(NULL);
   p = (*plist)->next;  // p chi nut dau dslk vong
   for(vitri = 0; vitri < i; vitri++)
      p = p->next;
   return(p);
}

// Tac vu push: them nut vao dau danh sach lien ket vong
void push(NODEPTR *plist, hoangtu x)
{
   NODEPTR p;
   p = getnode();
   p->info = x;
   if(empty(plist))
      *plist = p;
   else
      p->next = (*plist)->next;
   (*plist)->next = p;
}

// Tac vu insend: them nut vao cuoi danh sach lien ket vong
void insend(NODEPTR *plist, hoangtu x)
{
   NODEPTR p;
   p = getnode();
   p->info = x;
   if(empty(plist))  // truong hop danh sach lien ket vong dang bi rong
      p->next = p;
   else
   {
      p->next = (*plist)->next;
      (*plist)->next = p;
   }
   *plist = p;
}

/* Tac vu insafter: them nut sau nut p, khong dung tac vu nay de them
   nut vao cuoi danh sach lien ket vong */
void insafter(NODEPTR *plist, NODEPTR p, hoangtu x)
{
   NODEPTR q;
   if(p == NULL)
      printf("Nut p khong hien huu, khong them nut duoc");
   else
   {
      if(p == *plist)
         printf("Ban dung tac vu insend de them nut nay vao cuoi dslk vong");
      else
      {
         // Them nut q vao danh sach sau nut p
         q = getnode();
         q->info = x;
         q->next = p->next;
         p->next = q;
      }
   }
}

// Tac vu pop: xoa nut o dau danh sach lien ket vong
hoangtu pop(NODEPTR *plist)
{
   NODEPTR p;
   hoangtu x;
   if(empty(plist))
      printf("Danh sach bi rong");
   else
   {
      p = (*plist)->next;  // p la nut can xoa
      x = p->info;
      if(p == *plist)      // truong hop danh sach chi co mot nut
         *plist = NULL;
      else
         (*plist)->next = p->next;
      freenode(p);
      return(x);
   }
}

// Tac vu dellast: xoa nut o cuoi danh sach lien ket vong
hoangtu dellast(NODEPTR *plist)
{
   NODEPTR p;
   hoangtu x;
   if(empty(plist))  // truong hop danh sach bi rong
      printf("Danh sach bi rong");
   else
   {
      if((*plist)->next == *plist)   // truong hop chi co mot nut
      {
         p = *plist;
         x = p->info;
         *plist = NULL;
         freenode(p);
         return(x);
      }
      p = *plist;     // p la nut can xoa
      x = p->info;
      *plist = nodepointer(plist, listsize(plist)-2);
      (*plist)->next = p->next;
      freenode(p);
      return(x);
   }
}

/* Tac vu delafter: xoa nut sau nut p, khong dung tac vu nay de xoa
   nut o cuoi danh sach lien ket vong */
hoangtu delafter(NODEPTR *plist, NODEPTR p)
{
   NODEPTR q;
   hoangtu x;
   if(empty(plist))  // truong hop danh sach bi rong
      printf("Danh sach bi rong");
   else
   {
      if(p == NULL)
         printf("Nut p khong hien huu, khong xoa nut duoc");
      else
      {
         if(p == *plist) // truong hop p la nut cuoi cua danh sach
            printf("Nut p la nut cuoi, khong the xoa nut sau nut p duoc");
         else
         {
            if(p->next == *plist)     // xoa nut cuoi danh sach
               printf("Ban nen dung tac vu dellast de xoa nut cuoi danh sach");
            else
            {
               // xoa nut q trong danh sach sau nut p
               q = p->next;    // q la nut can xoa
               x = q->info;
               p->next = q->next;
               freenode(q);
               return(x);
            }
         }
      }
   }
}

// Tac vu traverse: duyet danh sach tu nut dau den nut cuoi
void traverse(NODEPTR *plist)
{
   NODEPTR p;
   if(*plist == NULL)
      printf("Danh sach bi rong");
   else
   {
      p = (*plist)->next;  // p chi nut dau
      printf("%s ", p->info.ten);
      p = p->next;         // p chi nut thu hai
      while(p != (*plist)->next)
      {
         printf("%s ", p->info.ten);
         p = p->next;
      }
   }
}

// Tac vu clearlist: xoa tat ca cac nut trong danh sach lien ket vong
void clearlist(NODEPTR *plist)
{
   while(*plist != NULL)
      pop(plist);
}

void main()
{
   NODEPTR plist, p;   /* plist la con tro chi nut cuoi danh sach lien
                          ket vong */
   hoangtu ht;
   int vitri, chucnang, n, i;

   clrscr();

   // khoi dong danh sach lien ket vong chua cac hoang tu
   initialize(&plist);

   do
   {
      // menu chinh cua chuong trinh
      printf("\n\nCHUONG TRINH KEN CHON PHO MA - DUNG GIAI THUAT JOSEPHUS\n");
      printf("   1: Khoi dong danh sach lien ket vong\n");
      printf("   2: Nhap hoang tu moi vao danh sach\n");
      printf("   3: Xem danh sach cac hoang tu\n");
      printf("   4: Ken chon pho ma bang giai thuat Josephus\n");
      printf("   0: Ket thuc chuong trinh\n");
      printf("Chuc nang ban chon: ");
      scanf("%d", &chucnang);
      switch(chucnang)
      {
	      case 1:
	      {
            clearlist(&plist);
	         break;
	      }
	      case 2:
	      {
	         printf("\nChon vi tri tren vong (0,1,2,...): ");
	         scanf("%d", &vitri);
	         printf("Ten hoang tu: ");
	         scanf("%s", &ht.ten);
	         if(vitri == 0)
	            push(&plist, ht);
	         else
	            if(vitri >= listsize(&plist))
		            insend(&plist, ht);
	            else  //insafter
		            insafter(&plist, nodepointer(&plist, vitri-1), ht);
	         break;
	      }
	      case 3:
	      {
	         printf("\nDanh sach cac hoang tu: ");
	         traverse(&plist);
	         break;
	      }
	      case 4:
	      {
	         printf("\nBuoc dem: ");
            scanf("%d", &n);
            p = plist;
            if(p == NULL)
               printf("Khong co hoang tu den du thi");
            else
            {
               /* vong lap xoa cac nut theo buoc dem, cuoi cung danh sach
                  lien ket vong chi con lai mot nut */
               while(plist != plist->next)
               {
                  // chuyen den nut p la nut truoc nut bi xoa
                  for(i = 1; i < n; i++)
                     p = p->next;

                  // xoa nut sau nut p
                  if(p->next == plist->next)  // xoa nut dau
                     pop(&plist);
                  else
                     if(p->next == plist)       // xoa nut cuoi
                        dellast(&plist);
                     else
                        delafter(&plist, p);      // xoa nut sau nut p
               }

               printf("\nHoang tu duoc chon lam pho ma la: %s", plist->info.ten);
               clearlist(&plist);
            }
            break;
	      }
      }
   } while(chucnang != 0);
   clearlist(&plist);
}