/*****************************************************************************
* Chuong trinh C8-1: Hash Tables, phuong phap noi ket truc tiep (direct      *
* chaining method) cai dat bang danh sach lien ket                           *
* Nguoi viet chuong trinh: Nguyen Hong Chuong                                *
* Ngay cap nhat moi nhat: 15/3/1999                                          *
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <alloc.h>

#define TRUE 1
#define FALSE 0
#define M 10

struct nodes
{
   int key;
   struct nodes *next;
};
typedef struct nodes *NODEPTR;

NODEPTR bucket[M];  // mang cac con tro chi nut dau cua cac bucket

// Tac vu getnode: cap phat mot nut cho bang bam
NODEPTR getnode(void)
{
   NODEPTR p;
   p = (NODEPTR)malloc(sizeof(struct nodes));
   return(p);
}

// Tac vu freenode: huy nut da cap phat
void freenode(NODEPTR p)
{
   free(p);
}

// Ham bam
int hashfunc(int key)
{
   return(key % M);
}

// Khoi dong cac bucket
void initbucket()
{
   int b;
   for(b = 0; b < M; b++)
      bucket[b] = NULL;
}

// Tac vu emptybucket: kiem tra bucket b co rong khong
int emptybucket(int b)
{
   return(bucket[b] == NULL ? TRUE : FALSE);
}

// Tac vu empty: kiem tra bang bam co rong khong
int empty()
{
   int b;
   for(b = 0; b < M; b++)
      if(bucket[b] != NULL)
	      return(FALSE);
   return(TRUE);
}

// Tac vu push: them nut moi vao dau bucket b
void push(int b, int x)
{
   NODEPTR p;
   p = getnode();
   p->key = x;
   p->next = bucket[b];
   bucket[b] = p;
}

// Tac vu pop: xoa nut o dau bucket b
int pop(int b)
{
   NODEPTR p;
   int k;
   if(emptybucket(b))
   {
      printf("\nBucket %d rong, khong xoa nut duoc", b);
      return(0);
   }
   p = bucket[b];                 // nut can xoa la nut dau Bucket b
   k = p->key;                    // k la noi dung nut bi xoa
   bucket[b] = p->next;
   freenode(p);
   return(k);
}

// Tac vu insafter: them nut moi vao bucket sau nut p
void insafter(NODEPTR p, int k)
{
   NODEPTR q;
   if(p == NULL)
      printf("khong them nut moi duoc");
   else
   {
      q = getnode();
      q->key = k;
      q->next = p->next;
      p->next = q;
   }
}

// Tac vu delafter: xoa nut trong bucket sau nut p
int delafter(NODEPTR p)
{
   NODEPTR q;
   int k;

   if(p == NULL || p->next == NULL)
   {
      printf("khong xoa nut duoc");
      return(0);
   }
   q = p->next;                                   // q chi nut can xoa
   k = q->key;                                    // k la noi dung nut bi xoa
   p->next = q->next;
   freenode(q);
   return(k);
}

// Tac vu place: tac vu nay chi su dung khi them nut vao bucket da co thu tu
void place(int b, int k)
{
   NODEPTR p, q;                              // q la nut truoc, p la nut sau
   q = NULL;
   for(p = bucket[b]; p != NULL && k > p->key; p = p->next)
      q = p;
   if(q == NULL)                                    // them nut vao dau bucket
      push(b, k);
   else                                             // them nut vao sau nut q
      insafter(q, k);
}

// Tac vu insert: them nut co khoa k vao bang bam
void insert(int k)
{
   int b;
	b = hashfunc(k);
	place(b, k);    // tac vu place cua danh sach lien ket
}

// Tac vu remove: xoa nut co khoa k trong bang bam
void remove(int k)
{
   int b;
   NODEPTR p, q;
	b = hashfunc(k);
	p = bucket[b];
	q = p;
	while(p != NULL && p->key != k)
	{
	   q = p;
	   p = p->next;
	}
	if(p == NULL)
	   printf("\n   Khong co nut co khoa %d", k);
	else
	   if(p == bucket[b])
		   pop(b);        // tac vu pop cua danh sach lien ket
	   else
		   delafter(q);   // tac vu delafter cua danh sach lien ket
}

// Tac vu clearbucket: xoa tat ca cac nut trong bucket b
void clearbucket(int b)
{
   NODEPTR p, q;                              // q la nut truoc, p la nut sau
   q = NULL;
   p = bucket[b];
   while(p != NULL)
   {
      q = p;
      p = p->next;
      freenode(q);
   }
   bucket[b] = NULL;                                   // khoi dong lai bucket b
}

// Tac vu clear: xoa tat ca cac nut trong bang bam
void clear()
{
   int b;
   for(b = 0; b < M; b++)
      clearbucket(b);
}

// Tac vu traversebucket: duyet bucket b
void traversebucket(int b)
{
   NODEPTR p;
   p = bucket[b];
   while(p != NULL)
   {
      printf("%3d ", p->key);
      p = p->next;
   }
}

// Tac vu traverse: duyet bang bam
void traverse()
{
   int b;
   for(b = 0; b < M; b++)
   {
      printf("\nBucket %d:", b);
      traversebucket(b);
   }
}

/* Tac vu search: tim kiem mot nut trong bang bam, neu khong tim thay ham nay
   tra ve tri -1, neu tim thay ham nay tra ve bucket co chua nut */
int search(int k)
{
   NODEPTR p;
   int b;
	b = hashfunc(k);
   p = bucket[b];
   while(k > p->key && p != NULL)
      p = p->next;
   if(p == NULL || k != p->key)          // khong tim thay
      return(-1);
   else                                  // tim thay
      return(b);
}

// chuong trinh chinh
void main()
{
   int b, key, i, n, chucnang;
   char c;

   clrscr();

   // khoi dong M bucket cua bang bam
   initbucket();

   do
   {
      // menu chinh cua chuong trinh
      printf("\n\nCac chuc nang cua chuong trinh:\n");
      printf("   1: Them mot nut vao bang bam\n");
      printf("   2: Them ngau nhien nhieu nut vao bang bam\n");
      printf("   3: Xoa nut trong bang bam\n");
      printf("   4: Xoa toan bo bang bam\n");
      printf("   5: Duyet bang bam\n");
      printf("   6: Tim kiem tren bang bam\n");
      printf("   0: Ket thuc chuong trinh\n");
      printf("\nChuc nang ban chon: ");
      scanf("%d", &chucnang);
      switch(chucnang)
      {
	      case 1:
	      {
	         printf("\nTHEM MOT NUT VAO BANG BAM");
	         printf("\n   Khoa cua nut moi: ");
	         scanf("%d", &key);
            insert(key);
	         break;
	      }
	      case 2:
	      {
	         printf("\nTHEM NGAU NHIEN NHIEU NUT VAO BANG BAM");
	         printf("\n   Ban muon them bao nhieu nut: ");
	         scanf("%d", &n);
	         for(i = 0; i < n; i++)
	         {
	            key = random(100);
               insert(key);
	         }
	         break;
	      }
	      case 3:
	      {
	         printf("\nXOA NUT TREN BANG BAM");
	         printf("\n   Khoa cua nut can xoa: ");
	         scanf("%d", &key);
            remove(key);
	         break;
	      }
	      case 4:
	      {
	         printf("\nXOA TOAN BO BANG BAM");
	         printf("\n   Ban co chac khong (c/k): ");
	         c = getche();
	         if(c == 'c' || c == 'C')
	            clear();
	         break;
	      }
	      case 5:
	      {
	         printf("\nDUYET BANG BAM");
	         traverse();
	         break;
	      }
	      case 6:
	      {
	         printf("\nTIM KIEM TREN BANG BAM");
	         printf("\n   Khoa can tim: ");
	         scanf("%d", &key);
            b = search(key);
	         if(b == -1)
	            printf("   Khong thay");
	         else
	            printf("   Tim thay trong bucket %d", b);
	         break;
	      }
      }
   } while(chucnang != 0);

   // Xoa tat ca cac nut tren bang bam
   clear();
}