/*****************************************************************************
* Chuong trinh C8-2: Hash Tables, dung phuong phap noi ket hop nhat          *
* (coalesced chaining method). Hash Table duoc cai dat bang danh sach ke,    *
* moi nut co truong next chi nut ke tiep khi co xung dot                     *
* Chuong trinh nay khong cai dat tac vu remove                               *
* Nguoi viet chuong trinh: Nguyen Hong Chuong                                *
* Ngay cap nhat moi nhat: 15/3/1999                                          *
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>

#define TRUE 1
#define FALSE 0

#define NULLKEY -1
#define M 10    /* M la so nut co tren bang bam, du de chua cac nut nhap vao
                   bang bam */

// Khai bao cau truc mot nut cua bang bam
struct node
{
   int key;   // khoa cua nut tren bang bam
   int next;  // con tro chi nut ke tiep khi co xung dot
};

// Khai bao bang bam
struct node hashtable[M];

int avail;    /* bien toan cuc chi nut trong o cuoi table duoc
			        cap phat khi co xung dot */
// Ham bam
int hashfunc(int key)
{
   return(key % M);
}

// Khoi dong bang bam
void initialize()
{
   int i;
   for(i = 0; i < M; i++)
   {
      hashtable[i].key = NULLKEY;
      hashtable[i].next = -1;
   }
   avail = M-1;    /* nut M-1 la nut o cuoi bang chuan bi cap phat neu co
                      xung dot */
}

// Tac vu empty: kiem tra bang bam co rong khong
int empty()
{
   int i;
   for(i = 0; i < M; i++)
      if(hashtable[i].key != NULLKEY)
	      return(FALSE);
   return(TRUE);
}

/* Tac vu search: tim kiem theo phuong phap tuyen tinh, neu khong tim
   thay ham nay tra ve tri M, neu tim thay ham nay tra ve dia chi tim thay */
int search(int k)
{
   int i;
   i = hashfunc(k);
   while(k != hashtable[i].key && i != -1)
      i = hashtable[i].next;
   if(k == hashtable[i].key)      // tim thay
      return(i);
   else                           // khong tim thay
      return(M);
}

/* Ham getempty: chon nut con trong phia cuoi hashtable de cap phat khi xay
   ra xung dot */
int getempty()
{
   while(hashtable[avail].key != NULLKEY)
      avail--;
   return(avail);
}

// Tac vu insert: them nut co khoa k vao bang bam
int insert(int k)
{
   int i,     // con tro lan theo danh sach lien ket chua cac nut bi xung dot
       j;     // dia chi nut trong duoc cap phat
   i = search(k);
   if(i != M)
   {
      printf("\n   khoa %d bi trung, khong them nut nay duoc", k);
      return(i);
   }
   i = hashfunc(k);
   while(hashtable[i].next >= 0)
      i = hashtable[i].next;
   if(hashtable[i].key == NULLKEY) // Neu nut i con trong thi cap phat
      j = i;
   else                            // Neu nut i la nut cuoi cua DSLK
   {
      j = getempty();
      if(j < 0)
      {
         printf("\n   Bang bam bi day, khong them nut co khoa %d duoc", k);
	      return(j);
      }
      else
	      hashtable[i].next = j;
   }
   hashtable[j].key = k;
   return(j);
}

// Tac vu viewtable: xem chi tiet bang bam
void viewtable()
{
   int i;
   for(i = 0; i < M; i++)
      printf("\ntable[%2d]: %4d %4d", i, hashtable[i].key, hashtable[i].next);
}

// chuong trinh chinh
void main()
{
   int i, n, p, q;
   int b, key, chucnang;
   char c;

   clrscr();

   // khoi dong bang bam
   initialize();

   do
   {
      // menu chinh cua chuong trinh
      printf("\n\nCac chuc nang cua chuong trinh:\n");
      printf("   1: Them nut moi vao bang bam\n");
      printf("   2: Them ngau nhien nhieu nut vao bang bam\n");
      printf("   3: Xoa toan bo bang bam\n");
      printf("   4: Xem chi tiet bang bam\n");
      printf("   5: Tim kiem tren bang bam\n");
      printf("   0: Ket thuc chuong trinh\n");
      printf("\nChuc nang ban chon: ");
      scanf("%d", &chucnang);
      switch(chucnang)
      {
	      case 1:
	      {
	         printf("\nTHEM NUT MOI VAO BANG BAM");
	         printf("\n   Khoa cua nut moi: ");
	         scanf("%d", &key);
	         insert(key);
	         break;
	      }
	      case 2:
	      {
	         printf("\nTHEM NGAU NHIEN NHIEU NUT VAO BANG BAM");
	         printf("\n   Ban muon them bao nhieu nut: ");
	         scanf("%d", &n);
	         for(i = 0; i < n; i++)
	         {
	            key = random(1000);
	            insert(key);
	         }
	         break;
	      }
	      case 3:
	      {
	         printf("\nXOA TOAN BO BANG BAM");
	         printf("\n   Ban co chac khong (c/k): ");
	         c = getche();
	         if(c == 'c' || c == 'C')
	            initialize();
	         break;
	      }
	      case 4:
	      {
	         printf("\nXEM CHI TIET BANG BAM");
	         viewtable();
	         break;
	      }
	      case 5:
	      {
	         printf("\nTIM KIEM TREN BANG BAM");
	         printf("\n   Khoa can tim: ");
	         scanf("%d", &key);
	         if(search(key) == M)
	            printf("   Khong thay");
	         else
	            printf("   Tim thay tai dia chi %d trong bang bam", search(key));
	         break;
	      }
      }
   } while(chucnang != 0);
}