/*****************************************************************************
* Chuong trinh C8-3: Bang bam, dung phuong phap do tuyen tinh (linear        *
* probing method). Bang bam duoc cai dat bang danh sach ke                   *
* Chuong trinh nay khong cai dat tac vu remove                               *
* Nguoi viet chuong trinh: Nguyen Hong Chuong                                *
* Ngay cap nhat moi nhat: 15/3/1999                                          *
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>

#define TRUE 1
#define FALSE 0
#define NULLKEY -1
#define M 10

// Khai bao cau truc mot nut cua bang bam
struct node
{
   int key;   // khoa cua nut tren bang bam
};

// Khai bao bang bam co M nut
struct node hashtable[M];

int N;        // bien toan cuc chi so nut hien co tren bang bam

// Ham bam
int hashfunc(int key)
{
   return(key % M);
}

// Khoi dong bang bam
void initialize()
{
   int i;
   for(i = 0; i < M; i++)
      hashtable[i].key = NULLKEY;
   N = 0;     // so nut hien co khoi dong bang 0
}

// Tac vu empty: kiem tra bang bam co rong khong
int empty()
{
   return(N == 0 ? TRUE : FALSE);
}

// Tac vu full: kiem tra bang bam da day chua
int full()
{
   return(N == M-1 ? TRUE : FALSE);
}

/* Tac vu search: tim kiem nut co khoa k tren bang bam, neu khong tim
   thay ham nay tra ve tri M, neu tim thay ham nay tra ve dia chi tim thay */
int search(int k)
{
   int i;
   i = hashfunc(k);
   while(hashtable[i].key != k && hashtable[i].key != NULLKEY)
   {
      // bam lai (theo phuong phap do tuyen tinh: hi(key) = (h(key)+i) % M )
      i = i+1;
      if(i >= M)
	      i = i-M;
   }
   if(hashtable[i].key == k)                        // tim thay
      return(i);
   else                                         // khong tim thay
      return(M);
}

// Tac vu insert: them nut co khoa k vao bang bam
int insert(int k)
{
   int i, j;
   if(full())
   {
      printf("\n   Bang bam bi day, khong them nut co khoa %d duoc", k);
      return(M);
   }
   i = hashfunc(k);
   while(hashtable[i].key != NULLKEY)
   {
      // bam lai (theo phuong phap do tuyen tinh)
      i++;
      if(i >= M)
	      i = i-M;
   }
   hashtable[i].key = k;
   N = N+1;
   return(i);
}

// Tac vu remove: xoa nut tai dia chi i tren bang bam
void remove(int i)
{
   int j, r, cont, a;
   cont = TRUE;
   do
   {
      hashtable[i].key = NULLKEY;
      j = i;
      do
      {
	      i = i+1;
	      if(i >= M)
	         i = i-M;
	      if(hashtable[i].key == NULLKEY)
	         cont = FALSE;
	      else
	      {
	         r = hashfunc(hashtable[i].key);
	         a = (j < r && r <= i) || (r <= i && i < j) || (i < j && j < r);
	      }
      } while (cont && a);
      if(cont)
	      hashtable[j].key = hashtable[i].key;
   } while(cont);
}

// Tac vu viewtable: xem chi tiet bang bam
void viewtable()
{
   int i;
   for(i = 0; i < M; i++)
      printf("\ntable[%2d]: %4d", i, hashtable[i].key);
}

// chuong trinh chinh
main()
{
   int i, n, p, q;
   int b, key, chucnang;
   char c;

   clrscr();

   // khoi tao bang bam
   initialize();

   do
   {
      // menu chinh cua chuong trinh
      printf("\n\nCac chuc nang cua chuong trinh:\n");
      printf("   1: Them nut moi vao bang bam\n");
      printf("   2: Them ngau nhien nhieu nut vao bang bam\n");
      printf("   3: Xoa nut tren bang bam\n");
      printf("   4: Xoa toan bo bang bam\n");
      printf("   5: Xem chi tiet bang bam\n");
      printf("   6: Tim kiem tren bang bam\n");
      printf("   0: Ket thuc chuong trinh\n");
      printf("\nChuc nang ban chon: ");
      scanf("%d", &chucnang);
      switch(chucnang)
      {
	      case 1:
	      {
	         printf("\nTHEM NUT MOI VAO BANG BAM");
	         printf("\n   Khoa cua nut moi: ");
	         scanf("%d", &key);
	         insert(key);
	         break;
	      }
	      case 2:
	      {
	         printf("\nTHEM NGAU NHIEN NHIEU NUT VAO BANG BAM");
	         printf("\n   Ban muon them bao nhieu nut: ");
	         scanf("%d", &n);
	         for(i = 0; i < n; i++)
	         {
	            key = random(1000);
	            insert(key);
	         }
	         break;
	      }
	      case 3:
	      {
	         printf("\nXOA NUT TREN BANG BAM");
	         printf("\n   Khoa cua nut can xoa: ");
	         scanf("%d", &key);
	         i = search(key);
	         if(i == M)
	            printf("Khong co nut voi khoa can xoa");
	         else
	         {
	            remove(i);
	            N--;
	         }
	         break;
	      }
	      case 4:
	      {
	         printf("\nXOA TOAN BO BANG BAM");
	         printf("\n   Ban co chac khong (c/k): ");
	         c = getche();
	         if(c == 'c' || c == 'C')
	            initialize();
	         break;
	      }
	      case 5:
	      {
	         printf("\nXEM CHI TIET BANG BAM");
	         viewtable();
	         break;
	      }
	      case 6:
	      {
	         printf("\nTIM KIEM TREN BANG BAM");
	         printf("\n   Khoa can tim: ");
	         scanf("%d", &key);
	         if(search(key) == M)
	            printf("   Khong thay");
	         else
	            printf("   Tim thay tai dia chi %d trong bang bam", search(key));
	         break;
	      }
      }
   } while(chucnang != 0);
   return(0);
}