/*****************************************************************************
* Chuong trinh C9-2: Chuong trinh minh hoa B-Tree to chuc o bo nho ngoai     *
* To chuc tap tin B-Tree chua cac mau tin (struct), moi mau tin la mot nut   *
* cua B-Tree:                                                                *
*     - Mau tin dau: danh rieng de chua header, luu cac thong tin nhu ptree  *
*       (dia chi cua nut goc), numnodes (so nut co tren B-Tree), depth (chieu*
*       sau cua B-Tree), ...                                                 *
*     - Cac mau tin sau: chua cac nut cua B-Tree, moi nut co cac field sau:  *
*           int numtrees;        // so nhanh cay con cua nut                 *
*           int key[ORDER-1];    // cac khoa cua nut                         *
*           int son[ORDER];      // cac dia chi cua cac nut con cua nut      *
* Nguoi viet chuong trinh: Nguyen Hong Chuong                                *
* Ngay cap nhat moi nhat: 15/3/1999                                          *
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <alloc.h>

#define ORDER 5    // bac cua B-Tree
#define Ndiv2 ORDER/2
#define TRUE 1
#define FALSE 0

// Khai bao cau truc mot nut cua B-Tree o bo nho ngoai
struct node
{
   int numtrees;        // so nhanh cay con cua nut
   int key[ORDER-1];    // cac khoa cua nut
   int son[ORDER];      // cac dia chi cua cac nut con cua nut
};

struct node pathnode[20];    /* mang chua cac nut tren duong di tu nut goc
                                den nut la */
int location[20];            // mang chua dia chi cua cac nut tren pathnode

int ptree;    // dia chi nut goc cua B-Tree

// Tac vu initialize: khoi dong B-Tree
void initialize()
{
   FILE *f;
   struct node header;
   header.numtrees = -1;
   header.key[0] = 0;
   f = fopen("c:\\btree.dat", "wb");
   fwrite(&header, sizeof(struct node), 1, f);
   fclose(f);
}

// Tac vu readheader: doc cac thong tin trong header cua tap tin B-Tree
void readheader()
{
   FILE *f;
   struct node header;
   f = fopen("c:\\btree.dat", "rb");
   fread(&header, sizeof(struct node), 1, f);
   ptree = header.numtrees;  // truy xuat dia chi nut goc cua B-Tree
   fclose(f);
}

// Tac vu makeroot: tao nut goc moi x cho B-Tree
int makeroot(struct node x)
{
   FILE *f;
   struct node temp;
   if((f = fopen("c:\\btree.dat","r+b")) == NULL)
   {
		printf("Khong mo tap tin chua B-Tree duoc\n");
		return(0);
   }

   // cap nhat thong tin trong header
   fread(&temp, sizeof(struct node), 1, f);
   temp.numtrees = temp.key[0];   // cap nhat lai dia chi nut goc moi
   temp.key[0]++;                 // tang so nut co trong B-Tree
   fseek(f, 0, SEEK_SET);
   fwrite(&temp, sizeof(struct node), 1, f);     // ghi lai header

   // append nut goc moi vao cuoi tap tin
   fseek(f, 0, SEEK_END);
   fwrite(&x, sizeof(struct node), 1, f);
   fclose(f);
   return(temp.key[0]-1);    // tra ve dia chi cua nut goc moi
}

/* Tac vu makenode: tao nut moi pathnode[i] cho B-Tree. Nut moi nay la nut
   nua phai cua lan tach nut */
int makenode(int i)
{
   FILE *f;
   struct node temp;
   if((f = fopen("c:\\btree.dat","r+b")) == NULL)
   {
		printf("Khong mo tap tin chua B-Tree duoc\n");
		return(0);
   }

   // cap nhat thong tin trong header
   fread(&temp, sizeof(struct node), 1, f);
   temp.key[0]++;  // tang so nut co trong B-Tree
   fseek(f, 0, SEEK_SET);
   fwrite(&temp, sizeof(struct node), 1, f);     // ghi lai header

   // append nut moi vao cuoi tap tin
   fseek(f, 0, SEEK_END);
   fwrite(&pathnode[i], sizeof(struct node), 1, f);
   fclose(f);
   return(temp.key[0]-1);    // tra ve dia chi cua nut moi
}

/* Tac vu access: Nap nut tai dia chi loc vao pathnode[i], nap dia chi loc
   vao location[i] */
void access(int i, int loc)
{
   FILE *f;
   if((f = fopen("c:\\btree.dat", "rb")) == NULL)
   {
		printf("Khong mo tap tin chua B-Tree duoc\n");
		return;
   }
   fseek(f, (loc+1)*sizeof(struct node), SEEK_SET);

   // Chep nut tai dia chi loc vao pathnode[i]
   fread(&pathnode[i], sizeof(struct node), 1, f);

   // Chep dia chi loc vao location[i]
   location[i] = loc;

   fclose(f);
}

/* Tac vu replace: cap nhat lai nut pathnode[i] tai dia chi loc cua tap tin
   B-Tree */
void replace(int i, int loc)
{
   FILE *f;
   if((f = fopen("c:\\btree.dat", "r+b")) == NULL)
   {
		printf("Khong mo tap tin chua B-Tree duoc\n\n");
		return;
   }
   fseek(f, (loc+1)*sizeof(struct node), SEEK_SET);
   fwrite(&pathnode[i], sizeof(struct node), 1, f);
   fclose(f);
}

// Tac vu viewnodes: Xem chi tiet noi dung tung nut trong tap tin B-Tree
void viewnodes()
{
   FILE *f;
   struct node x;
   int i=0, j, numnodes;
   f = fopen("c:\\btree.dat", "rb");

   // Doc va hien thi header
   fread(&x, sizeof(struct node), 1, f);
   numnodes = x.key[0];
   printf("Dia chi nut goc: %d    So nut tren B-Tree: %d",
                             x.numtrees,               x.key[0]);

   // Doc va hien thi thong tin cua tung nut trong B-Tree
   while(i < numnodes)
   {
      printf("\nDia chi %d:", i);
      fread(&x, sizeof(struct node), 1, f);
      printf("   bac: %d", x.numtrees);
      printf("   khoa va nut con: ");
      for(j = 0; j < x.numtrees-1; j++)
         printf("%d ", x.key[j]);
      printf("  ");
      for(j = 0; j < x.numtrees; j++)
         printf("%d ", x.son[j]);
      i++;
   }
   fclose(f);
}

/* Tac vu nodesearch: tim vi tri trong nut pathnode[p] (tra ve vi tri cua
   khoa bat dau lon hon hay bang k). Truong hop k lon hon tat ca cac khoa
   thi tra ve pathnode[p].numtrees-1 */
int nodesearch(int p, int k)
{
   int i;
   for(i = 0; i < pathnode[p].numtrees-1 && pathnode[p].key[i] < k; i++)
      ;
   return(i);
}

/* Tac vu search: tim khoa k tren B-Tree
   Neu co:
      - Bien found tra ve gia tri TRUE
      - Ham search() tra ve nut vi tri j trong pathnode co chua khoa k
      - Bien position tra ve vi tri cua khoa k co tren nut nay
   Neu khong co:
      - Bien found tra ve gia tri FALSE
      - Ham search() tra ve nut vi tri cuoi trong pathnode la nut co the them
        khoa k vao
      - Bien position tra ve vi tri cua khoa k neu them no vao nut nay */
int search(int k, int *pposition, int *pfound)
{
   int p, i, j;
   p = ptree; // p xuat phat tu nut goc cua B-Tree
   j = -1;    // j la chi so tren pathnode
   while(p != -1)
   {
      j++;
      access(j, p);
		i = nodesearch(j, k);
		if(i < pathnode[j].numtrees-1 && k == pathnode[j].key[i])
		{
			*pfound = TRUE;
			*pposition = i;
			return j;
		}
      p = pathnode[j].son[i];     // p chuyen xuong nut con
   }
   /* truong hop khong co khoa k tren B-Tree, luc nay da hinh thanh mang
      pathnode[] chua cac nut tren duong di tu nut goc den nut la */
   *pfound = FALSE;
   *pposition = i;
   return j;  // tra ve nut cuoi trong pathnode
}

/* Tac vu insnode: chen khoa newkey vao vi tri pos cua nut chua day
   pathnode[p], va chen nhanh cay con newnode vao vi tri ben phai cua khoa
   newkey */
void insnode(int p, int newkey, int newnode, int pos)
{
   int i;

   /* doi cac nhanh cay con va cac khoa tu vi tri pos tro ve sau xuong mot
      vi tri */
   for(i = pathnode[p].numtrees-1; i >= pos+1; i--)
   {
      pathnode[p].son[i+1] = pathnode[p].son[i];
      pathnode[p].key[i] = pathnode[p].key[i-1];
   }

   // Gan khoa newkey vao vi tri pos
   pathnode[p].key[pos] = newkey;

   // Gan nhanh newnode la nhanh cay con ben phai cua khoa newkey
   pathnode[p].son[pos+1] = newnode;

   // Tang so nhanh cay con  cua nut p len 1
   pathnode[p].numtrees += 1;
}

/* Tac vu copy: chep cac khoa (va nhanh cay con) tu vi tri first den vi tri
   last cua nut pathnode[nd] (nut nua trai) sang nut pathnode[nd1] (nut nua
   phai) */
void copy(int nd, int first, int last, int nd1)
{
   int i;

   // copy cac khoa tu nut pathnode[nd] qua pathnode[nd1]
   for(i = first; i <= last; i++)
      pathnode[nd1].key[i-first] = pathnode[nd].key[i];

   // copy cac nhanh cay con tu nut pathnode[nd] qua pathnode[nd1]
   for(i = first; i <= last+1; i++)
      pathnode[nd1].son[i-first] = pathnode[nd].son[i];

   pathnode[nd1].numtrees = last-first+2;   /* so nhanh cay con cua nut
                                               pathnode[nd1] */
}

/* Tac vu split: tach nut day pathnode[nd] thanh hai nut: nut nua trai
   pathnode[nd] va nut nua phai nd2 */
void split(int nd, int newkey, int newnode, int pos, int *pnd2, int *pmidkey)
{
   // truong hop chen newkey va newnode vao nut nua phai
   if(pos > Ndiv2)
   {
      copy(nd, Ndiv2+1, ORDER-2, nd+1); //dung pathnode[nd+1] luu nut nua phai
      insnode(nd+1, newkey, newnode, pos-Ndiv2-1);
      pathnode[nd].numtrees = Ndiv2+1;   /* so nhanh cay con con lai cua
                                            nut nua trai */
      *pmidkey = pathnode[nd].key[Ndiv2];
   }

   // truong hop newkey la midkey
   if(pos == Ndiv2)
   {
      copy(nd, Ndiv2, ORDER-2, nd+1);   //dung pathnode[nd+1] luu nut nua phai
      pathnode[nd].numtrees = Ndiv2+1;   /* so nhanh cay con con lai cua nut
                                            nua trai */
      // Dieu chinh lai nut con dau tien cua nut nua phai
      pathnode[nd+1].son[0] = newnode;
      *pmidkey = newkey;
   }

   // truong hop chen newkey va newnode vao nut nua trai
   if(pos < Ndiv2)
   {
      copy(nd, Ndiv2, ORDER-2, nd+1);   //dung pathnode[nd+1] luu nut nua phai
      pathnode[nd].numtrees = Ndiv2;     /* so nhanh cay con con lai cua nut
                                            nua trai */
      *pmidkey = pathnode[nd].key[Ndiv2-1];
      insnode(nd, newkey, newnode, pos);
   }

   /* cap nhat lai nut nua trai pathnode[nd] va tao nut nua phai vao tap tin
      B-Tree */
   replace(nd, location[nd]);
   *pnd2 = makenode(nd+1);
}

// Tac vu insert: Them khoa k vao vi tri position cua nut la pathnode[s].
void insert(int s, int k, int position)
{
   struct node x;
   int nd, nd2, newkey, newnode, pos, midkey, i;

   // Khoi dong cac tri truoc khi vao vong lap tach cac nut day pathnode[nd]
   nd = s;
   newkey = k;
   newnode = -1;
   pos = position;

   // Vong lap tach cac nut day pathnode[nd]
   while(nd != 0 && pathnode[nd].numtrees == ORDER)
   {
      split(nd, newkey, newnode, pos, &nd2, &midkey);
      // Gan lai cac tri sau lan tach nut truoc
      newkey = midkey;
      newnode = nd2;
      pos = nodesearch(nd-1, midkey);
      nd--;
   }

   /* Truong hop nut pathnode[nd] chua day va pathnode[nd] khong phai la nut
      goc */
   if(pathnode[nd].numtrees < ORDER)
   {
      // chen newkey va newnode tai vi tri pos cua nut pathnode[nd]
      insnode(nd, newkey, newnode, pos);
      replace(nd, location[nd]);
      return;
   }

   /* Truong hop nut pathnode[nd] la nut goc bi day, tach nut goc nay va tao
      nut goc moi */
   split(nd, newkey, newnode, pos, &nd2, &midkey);
   x.numtrees = 2;
   x.key[0] = midkey;
   // khoi dong dia chi cac nut con cua nut goc moi
   for(i = 0; i < ORDER; i++)
      x.son[i] = -1;
   // Gan lai hai nhanh cay con cua nut goc moi
   x.son[0] = location[nd];
   x.son[1] = nd2;
   ptree = makeroot(x);
}

// Chuong trinh chinh
void main()
{
   struct node x;
   int s, i, n, k, pos, timthay, chucnang;
   char c;

   clrscr();
   readheader();   // Doc cac thong tin trong header cua tap tin B-Tree

   do
   {
      // Menu chinh cua chuong trinh
      printf("\n\nCHUONG TRINH MINH HOA B-TREE TO CHUC O BO NHO NGOAI");
      printf("\n\nCac chuc nang cua chuong trinh:\n");
      printf("   1: Them mot khoa\n");
      printf("   2: Them ngau nhien nhieu khoa\n");
      printf("   3: Xem tap tin B-Tree\n");
      printf("   4: Khoi dong B-Tree\n");
      printf("   5: Tim kiem\n");
      printf("   0: Ket thuc chuong trinh\n");
      printf("Chuc nang ban chon: ");
      scanf("%d", &chucnang);
      switch(chucnang)
      {
	      case 1:
	      {
	         printf("\nNoi dung khoa moi: ");
	         scanf("%d", &k);

	         // Truong hop B-Tree bi rong thi tao nut goc
	         if(ptree == -1)
            {
               x.numtrees = 2;
               x.key[0] = k;
               for(i = 0; i < ORDER; i++)
                  x.son[i] = -1;
	            ptree = makeroot(x);
            }
	         else
	         {
	            s = search(k, &pos, &timthay);
	            if(timthay == TRUE)
		            printf("\nBi trung khoa, khong them khoa %d vao B-Tree duoc", k);
	            else
		            insert(s, k, pos); /* them khoa k vao nut la pathnode[s]
                                        tai vi tri pos */
	         }
	         break;
	      }
	      case 2:
	      {
	         printf("\nBan muon them vao bao nhieu khoa: ");
	         scanf("%d", &n);
	         for(i = 0; i < n; i++)
	         {
	            k = random(10000);
	            s = search(k, &pos, &timthay);
	            if(timthay == TRUE)
		            printf("\nBi trung khoa, khong them khoa %d vao B-Tree duoc", k);
	            else
		            insert(s, k, pos);
	         }
	         printf("\nDa them vao B-Tree %d so ngau nhien", n);
	         break;
	      }
	      case 3:
	      {
	         printf("\nXem tap tin B-Tree\n");
	         viewnodes();
	         break;
	      }
	      case 4:
	      {
            initialize();
            ptree = -1;
	         break;
	      }
	      case 5:
	      {
	         printf("\nKhoa can tim: ");
	         scanf("%d", &k);
	         s = search(k, &pos, &timthay);
	         if(timthay == TRUE)
	            printf("Tim thay tai vi tri %d cua nut co dia chi %d", pos, location[s]);
	         else
	            printf("Khong tim thay");
	         break;
	      }
      }
   } while(chucnang != 0);
}